import torch
import pytest
from PoPE_pytorch import PoPE, AxialPoPE, compute_attn_similarity, flash_attn_with_pope

def test_readme_usage():
    # define pope
    pope = PoPE(64, heads = 8)

    # pass in sequence length
    pos_emb = pope(1024)

    # queries and keys in attention
    q = torch.randn(1, 8, 1024, 64)
    k = torch.randn(1, 8, 1024, 64)

    # training
    rotated_q, rotated_k = pope.apply_pope_to_qk(pos_emb, q, k)
    assert rotated_q.shape == (1, 8, 1024, 128)
    assert rotated_k.shape == (1, 8, 1024, 128)

    # inference
    rotated_q, rotated_k = pope.apply_pope_to_qk(pos_emb, q[..., -1:, :], k)
    assert rotated_q.shape == (1, 8, 1, 128)
    assert rotated_k.shape == (1, 8, 1024, 128)

def test_readme_axial_pope():
    # axial pope for images (e.g. 32x32)
    # split 64 dim into 32 (x) and 32 (y)
    pope = AxialPoPE(
        dim = 64,
        heads = 8,
        axial_dims = (32, 32)
    )

    pos_emb = pope((32, 32)) # (1024, 64) frequencies
    assert pos_emb.freqs.shape == (1024, 64)

    # for video (e.g. 8 frames, 16x16 frames)
    # split 96 dim into 32 (t), 32 (x), 32 (y)
    pope_video = AxialPoPE(
        dim = 96,
        heads = 8,
        axial_dims = (32, 32, 32)
    )

    pos_emb_video = pope_video((8, 16, 16)) # (2048, 96) frequencies
    assert pos_emb_video.freqs.shape == (2048, 96)

    # queries and keys
    q = torch.randn(1, 8, 2048, 96)
    k = torch.randn(1, 8, 2048, 96)

    rotated_q, rotated_k = AxialPoPE.apply_pope_to_qk(pos_emb_video, q, k)
    assert rotated_q.shape == (1, 8, 2048, 192)

@pytest.mark.skipif(not torch.cuda.is_available(), reason = 'CUDA must be available')
def test_readme_fused_similarity():
    # define pope
    pope = PoPE(dim = 64, heads = 8).cuda()

    # get rotations
    pos_emb = pope(1024)

    # queries and keys
    q = torch.randn(1, 8, 1024, 64).cuda()
    k = torch.randn(1, 8, 1024, 64).cuda()

    # fused attention similarity, avoiding expanding 64 to 128
    sim = compute_attn_similarity(q, k, pos_emb) # (1, 8, 1024, 1024)
    assert sim.shape == (1, 8, 1024, 1024)

    attn = sim.softmax(dim = -1) # the usual in attention..

@pytest.mark.skipif(not torch.cuda.is_available(), reason = 'CUDA must be available')
def test_readme_fused_flash_attn():
    # pope
    pope = PoPE(dim = 32, heads = 8).cuda()

    # queries, keys, values for attention
    q = torch.randn(2, 8, 1024, 64).cuda()
    k = torch.randn(2, 8, 1024, 64).cuda()
    v = torch.randn(2, 8, 1024, 64).cuda()

    pos_emb = pope(1024)

    mask = torch.ones((2, 1024)).bool().cuda()

    out = flash_attn_with_pope(q, k, v, pos_emb = pos_emb, causal = True, mask = mask)

    assert out.shape == (2, 8, 1024, 64)



from PoPE_pytorch.pope import (
    PoPE,
    apply_pope_to_qk
)
from PoPE_pytorch.attention import (
    compute_attn_similarity,
    flash_attn_with_pope
)
from PoPE_pytorch.pope_nd import (
    AxialPoPE
)



import torch
import pytest
from PoPE_pytorch import PoPE
from PoPE_pytorch.attention import flash_attn_with_pope, compute_attn_similarity
from einops import rearrange

def exists(v):
    return v is not None

@pytest.mark.skipif(not torch.cuda.is_available(), reason = 'CUDA must be available')
@pytest.mark.parametrize('dtype', [torch.float16, torch.bfloat16, torch.float32])
@pytest.mark.parametrize('causal', [True, False])
@pytest.mark.parametrize('seq_len_q', [1, 256])
@pytest.mark.parametrize('has_mask', [True, False])
def test_fused_vs_manual(dtype, causal, seq_len_q, has_mask):
    device = 'cuda'
    
    batch, heads, dim = 2, 4, 64
    seq_len_k = 256
    rotate_dim = 32
    
    torch.manual_seed(42)
    
    q = torch.randn(batch, seq_len_q, heads, dim, device = device, dtype = dtype, requires_grad = True)
    k = torch.randn(batch, seq_len_k, heads, dim, device = device, dtype = dtype, requires_grad = True)
    v = torch.randn(batch, seq_len_k, heads, dim, device = device, dtype = dtype, requires_grad = True)
    
    pope_module = PoPE(rotate_dim, heads = heads).to(device)
    pope = pope_module(seq_len_k)
    
    do = torch.randn(batch, seq_len_q, heads, dim, device = device, dtype = dtype)

    key_pad_mask = None
    if has_mask:
        key_pad_mask = torch.ones((batch, seq_len_k), device = device, dtype = torch.bool)
        key_pad_mask[:, seq_len_k // 2:] = False

    # 1. Manual Path
    qc1, kc1, vc1 = q.clone().detach().requires_grad_(True), k.clone().detach().requires_grad_(True), v.clone().detach().requires_grad_(True)
    pc1 = [t.clone().detach().requires_grad_(True) for t in pope]
    
    out_manual = flash_attn_with_pope(qc1, kc1, vc1, pos_emb = pc1, mask = key_pad_mask, causal = causal, fused = False, head_dimension_at_first = False)
    out_manual.backward(do)
    
    # 2. Fused Path
    qc2, kc2, vc2 = q.clone().detach().requires_grad_(True), k.clone().detach().requires_grad_(True), v.clone().detach().requires_grad_(True)
    pc2 = [t.clone().detach().requires_grad_(True) for t in pope]
    
    out_fused = flash_attn_with_pope(qc2, kc2, vc2, pos_emb = pc2, mask = key_pad_mask, causal = causal, fused = True, head_dimension_at_first = False)
    out_fused.backward(do)
    
    # check parity
    def get_max_diff(a, b):
        return (a - b).abs().max().item()

    diff_out = get_max_diff(out_manual, out_fused)
    diff_dq = get_max_diff(qc1.grad, qc2.grad)
    diff_dk = get_max_diff(kc1.grad, kc2.grad)
    diff_dv = get_max_diff(vc1.grad, vc2.grad)
    
    print(f"\n[{dtype}, causal={causal}] Out Diff: {diff_out:.6f}, dQ Diff: {diff_dq:.6f}, dK Diff: {diff_dk:.6f}, dV Diff: {diff_dv:.6f}")

    atol = 5e-2 if dtype != torch.float32 else 2e-2
    
    assert torch.allclose(out_manual, out_fused, atol = atol), f"Out diff too large: {diff_out}"
    assert torch.allclose(qc1.grad, qc2.grad, atol = atol), f"dQ diff too large: {diff_dq}"
    assert torch.allclose(kc1.grad, kc2.grad, atol = atol), f"dK diff too large: {diff_dk}"
    assert torch.allclose(vc1.grad, vc2.grad, atol = atol), f"dV diff too large: {diff_dv}"
    
    if exists(pc1[0].grad) and exists(pc2[0].grad):
        diff_df = get_max_diff(pc1[0].grad, pc2[0].grad)
        assert torch.allclose(pc1[0].grad, pc2[0].grad, atol = atol), f"dfreqs diff too large: {diff_df}"

    if exists(pc1[1].grad) and exists(pc2[1].grad):
        diff_dpb = get_max_diff(pc1[1].grad, pc2[1].grad)
        assert torch.allclose(pc1[1].grad, pc2[1].grad, atol = atol), f"dpope_bias diff too large: {diff_dpb}"

@pytest.mark.skipif(not torch.cuda.is_available(), reason = 'CUDA must be available')
@pytest.mark.parametrize('seq_len_q', [1, 1024])
def test_compute_attn_similarity(seq_len_q):
    device = 'cuda'
    batch, heads, seq_len_k, dim = 1, 8, 1024, 64
    rotate_dim = 32

    q = torch.randn(batch, seq_len_q, heads, dim, device = device)
    k = torch.randn(batch, seq_len_k, heads, dim, device = device)
    v = torch.randn(batch, seq_len_k, heads, dim, device = device)

    pope_module = PoPE(rotate_dim, heads = heads).to(device)
    pope = pope_module(seq_len_k)

    # 1. Compute similarity then aggregate
    # We pass head_dimension_at_first = False because our q, k are (b, n, h, d)
    sim = compute_attn_similarity(q, k, pope, head_dimension_at_first = False)
    
    softmax_scale = dim ** -0.5
    attn = (sim * softmax_scale).softmax(dim = -1)
    
    # head-first for the einsum (b h n d)
    v_head_first = rearrange(v, 'b n h d -> b h n d')
    out_similarity = torch.einsum('b h i j, b h j d -> b h i d', attn, v_head_first)
    out_similarity = rearrange(out_similarity, 'b h n d -> b n h d')

    # 2. Use complete flash_attn_with_pope
    out_attention = flash_attn_with_pope(q, k, v, pos_emb = pope, head_dimension_at_first = False)

    # 3. Validate parity
    diff = (out_similarity - out_attention).abs().max().item()
    print(f"\n[seq_len_q={seq_len_q}] Similarity vs Attention Diff: {diff:.6f}")
    assert diff < 5e-4



import torch
import torch.nn.functional as F

from PoPE_pytorch.pope import apply_pope_to_qk

from einops import rearrange, repeat

# helpers

def exists(v):
    return v is not None

def default(v, d):
    return v if exists(v) else d

def divisible_by(num, den):
    return (num % den) == 0

# triton available

try:
    from .triton_pope import triton_compute_qk_similarity
    from .triton_pope_flash_attn import flash_attn
    TRITON_AVAILABLE = True
except ImportError:
    TRITON_AVAILABLE = False

# functions

def compute_attn_similarity_non_fused(
    q,
    k,
    pope,
    head_dimension_at_first = True
):
    if not head_dimension_at_first:
        q = rearrange(q, 'b n h d -> b h n d')
        k = rearrange(k, 'b n h d -> b h n d')
    
    q, k = apply_pope_to_qk(pope, q, k, to_magnitude = F.softplus)

    # group query attention support

    groups = q.shape[1] // k.shape[1]
    k = repeat(k, 'b h ... -> b (g h) ...', g = groups)

    return torch.einsum('b h i d, b h j d -> b h i j', q, k)

def compute_attn_similarity(
    q,
    k,
    pope,
    allow_tf32 = True,
    head_dimension_at_first = True
):
    assert divisible_by(q.shape[1 if head_dimension_at_first else 2], k.shape[1 if head_dimension_at_first else 2])

    freqs, bias = pope
    head_dim = q.shape[-1]

    assert head_dim in {32, 48, 64, 128, 256}, f"head_dim {head_dim} not in common sizes"

    is_cuda = q.is_cuda and k.is_cuda and freqs.is_cuda and bias.is_cuda

    if TRITON_AVAILABLE and is_cuda:
        if not head_dimension_at_first:
            q = rearrange(q, 'b n h d -> b h n d')
            k = rearrange(k, 'b n h d -> b h n d')

        rotate_dim = freqs.shape[-1]
        return triton_compute_qk_similarity(q, k, freqs, bias, rotate_dim, allow_tf32 = allow_tf32)

    return compute_attn_similarity_non_fused(q, k, pope, head_dimension_at_first = head_dimension_at_first)

def flash_attn_with_pope(
    q,
    k,
    v,
    pos_emb = None,
    mask = None,
    causal = False,
    softmax_scale = None,
    fused = None,
    head_dimension_at_first = True
):
    fused = default(fused, TRITON_AVAILABLE and q.is_cuda)

    softmax_scale = default(softmax_scale, q.shape[-1] ** -0.5)

    if fused:
        # fused kernel expects (batch, seq, heads, dim)

        if head_dimension_at_first:
            q = rearrange(q, 'b h n d -> b n h d')
            k = rearrange(k, 'b h n d -> b n h d')
            v = rearrange(v, 'b h n d -> b n h d')

        freqs, bias = pos_emb
        out = flash_attn(q, k, v, freqs = freqs, pope_bias = bias, mask = mask, causal = causal, softmax_scale = softmax_scale)

        if head_dimension_at_first:
            out = rearrange(out, 'b n h d -> b h n d')

        return out

    # non-fused manual path
    # standardize to (batch, heads, seq, dim)

    if not head_dimension_at_first:
        q = rearrange(q, 'b n h d -> b h n d')
        k = rearrange(k, 'b n h d -> b h n d')
        v = rearrange(v, 'b n h d -> b h n d')

    q, k = apply_pope_to_qk(pos_emb, q, k, to_magnitude = F.softplus)
    
    # group query attention support

    groups = q.shape[1] // k.shape[1]
    k = repeat(k, 'b h ... -> b (g h) ...', g = groups)
    v = repeat(v, 'b h ... -> b (g h) ...', g = groups)

    # manual attention path using SDPA
    # ensure dtypes match for SDPA (apply_pope_to_qk might have upcasted to float32)

    v_dtype = v.dtype
    if q.dtype != v.dtype:
        v = v.to(q.dtype)

    attn_mask = None
    if exists(mask):
        attn_mask = rearrange(mask, 'b j -> b 1 1 j')

    out = F.scaled_dot_product_attention(
        q, k, v,
        attn_mask = attn_mask,
        is_causal = causal,
        scale = softmax_scale
    )

    out = out.to(v_dtype)

    if not head_dimension_at_first:
        out = rearrange(out, 'b h n d -> b n h d')

    return out



from math import ceil, sqrt
from functools import partial

import torch
from torch import Tensor
from torch.autograd import Function

from einops import repeat, rearrange

import triton
import triton.language as tl

# helpers

def exists(v):
    return v is not None

def default(val, d):
    return val if exists(val) else d

# kernels

@triton.heuristics({
    "EVEN_M": lambda args: args["seqlen_q"] % args["BM"] == 0,
    "EVEN_N": lambda args: args["seqlen_k"] % args["BN"] == 0,
    "EVEN_HEADDIM": lambda args: args["headdim"] == args["BLOCK_HEADDIM"],
})
@triton.jit
def _fwd_kernel(
    Q, K, V, Freqs, PopeBias, Out, Lse, Mask,
    softmax_scale, SQB, SQH, SQM, SKB, SKH, SKN, SVB, SVH, SVN, 
    SFB, SFH, SFI, SPBH, SOB, SOH, SOM, SKMB, SKMN,
    n_heads, seqlen_q, seqlen_k, headdim, rotate_dim,
    HAS_POPE: tl.constexpr, IS_CAUSAL: tl.constexpr, HAS_MASK: tl.constexpr,
    BLOCK_HEADDIM: tl.constexpr, EVEN_M: tl.constexpr, EVEN_N: tl.constexpr, EVEN_HEADDIM: tl.constexpr,
    BM: tl.constexpr, BN: tl.constexpr,
):
    bh = tl.program_id(1)
    b, h = bh // n_heads, bh % n_heads
    sm = tl.program_id(0)

    om, on, od = sm * BM + tl.arange(0, BM), tl.arange(0, BN), tl.arange(0, BLOCK_HEADDIM)
    mm, md, mr = om < seqlen_q, od < headdim, od < rotate_dim

    qp = Q + b * SQB + h * SQH + om[:, None] * SQM + od[None, :]

    if EVEN_M & EVEN_HEADDIM:
        q = tl.load(qp)
    else:
        q = tl.load(qp, mask=mm[:, None] & md[None, :], other=0.0)

    if HAS_POPE:
        q = tl.where(mr[None, :], tl.where(q > 20., q, tl.log(1.0 + tl.exp(q.to(tl.float32))).to(q.dtype)), q)
        q_off = seqlen_k - seqlen_q
        fqp = Freqs + b * SFB + h * SFH + (q_off + om[:, None]) * SFI + od[None, :]
        fq = tl.load(fqp, mask=mm[:, None] & mr[None, :], other=0.0).to(tl.float32)
        qc, qs = tl.where(mr[None, :], q * tl.cos(fq).to(q.dtype), q), tl.where(mr[None, :], q * tl.sin(fq).to(q.dtype), 0.0)
    else:
        qc, qs = q, None

    m_i, l_i = tl.zeros([BM], tl.float32) - float("inf"), tl.zeros([BM], tl.float32)
    acc = tl.zeros([BM, BLOCK_HEADDIM], tl.float32)

    en = seqlen_k if not IS_CAUSAL else tl.minimum((sm + 1) * BM, seqlen_k)

    for sn in range(0, en, BN):
        cn = sn + on
        mn = cn < seqlen_k

        kp = K + b * SKB + h * SKH + cn[:, None] * SKN + od[None, :]
        k = tl.load(kp, mask=mn[:, None] & md[None, :], other=0.0) if not (EVEN_N & EVEN_HEADDIM) else tl.load(kp)

        if HAS_POPE:
            k = tl.where(mr[None, :], tl.where(k > 20., k, tl.log(1.0 + tl.exp(k.to(tl.float32))).to(k.dtype)), k)
            fkp = Freqs + b * SFB + h * SFH + cn[:, None] * SFI + od[None, :]
            fk = tl.load(fkp, mask=mn[:, None] & mr[None, :], other=0.0)
            pb = tl.load(PopeBias + h * SPBH + od, mask=mr, other=0.0)
            tk = (fk + pb[None, :]).to(tl.float32)
            kc, ks = tl.where(mr[None, :], k * tl.cos(tk).to(k.dtype), k), tl.where(mr[None, :], k * tl.sin(tk).to(k.dtype), 0.0)
            qk = tl.dot(qc, tl.trans(kc)) + tl.dot(qs, tl.trans(ks))
        else:
            qk = tl.dot(qc, tl.trans(k))

        qk *= softmax_scale

        if IS_CAUSAL:
            qk += tl.where(om[:, None] >= cn[None, :], 0, float("-inf"))

        if HAS_MASK:
            mask = tl.load(Mask + b * SKMB + cn * SKMN, mask=mn, other=False)
            qk += tl.where(mask[None, :], 0, float("-inf"))

        if not EVEN_N:
            qk += tl.where(mn[None, :], 0, float("-inf"))

        mij = tl.max(qk, 1)
        p = tl.exp(qk - tl.where(mij == float("-inf"), 0.0, mij)[:, None])
        p = tl.where(mij[:, None] == float("-inf"), 0.0, p)
        lij = tl.sum(p, 1)

        mi_new = tl.maximum(m_i, mij)
        mi_safe = tl.where(mi_new == float("-inf"), 0.0, mi_new)
        al, be = tl.exp(m_i - mi_safe), tl.exp(mij - mi_safe)

        acc *= al[:, None]

        vp = V + b * SVB + h * SVH + cn[:, None] * SVN + od[None, :]
        v = tl.load(vp, mask=mn[:, None] & md[None, :], other=0.0) if not (EVEN_N & EVEN_HEADDIM) else tl.load(vp)
        acc += tl.dot(p.to(v.dtype), v) * be[:, None]

        l_i = l_i * al + lij * be
        m_i = mi_new

    acc /= tl.where(l_i == 0.0, 1.0, l_i)[:, None]

    tl.store(Out + b * SOB + h * SOH + om[:, None] * SOM + od[None, :], acc.to(Out.dtype.element_ty), mask=mm[:, None] & md[None, :])
    tl.store(Lse + bh * seqlen_q + om, m_i + tl.log(l_i), mask=mm)

@triton.jit
def _bwd_preprocess(Out, DO, Delta, SOB, SOH, SOM, SDB, SDH, SDM, n_heads, seqlen_q, d, BM: tl.constexpr, BLOCK_D: tl.constexpr):
    hb = tl.program_id(1)
    b, h = hb // n_heads, hb % n_heads
    m = tl.program_id(0) * BM + tl.arange(0, BM)
    dd = tl.arange(0, BLOCK_D)
    mask = (m < seqlen_q)[:, None] & (dd < d)[None, :]

    o = tl.load(Out + b * SOB + h * SOH + m[:, None] * SOM + dd[None, :], mask = mask, other = 0.0).to(tl.float32)
    do = tl.load(DO + b * SDB + h * SDH + m[:, None] * SDM + dd[None, :], mask = mask, other = 0.0).to(tl.float32)

    tl.store(Delta + hb * seqlen_q + m, tl.sum(o * do, 1), mask = m < seqlen_q)

@triton.heuristics({
    "EVEN_M": lambda args: args["seqlen_q"] % args["BM"] == 0,
    "EVEN_N": lambda args: args["seqlen_k"] % args["BN"] == 0,
    "EVEN_HEADDIM": lambda args: args["headdim"] == args["BLOCK_HEADDIM"],
})
@triton.jit
def _bwd_kernel(
    Q, K, V, Freqs, PopeBias, DO, DQ, DK, DV, DFreqs, DPopeBias, Lse, Delta, Mask,
    softmax_scale, SQB, SQH, SQM, SKB, SKH, SKN, SVB, SVH, SVN, SFB, SFH, SFI, SPBH,
    SDB, SDH, SDM, SDQB, SDQH, SDQM, SDKB, SDKH, SDKN, SDVB, SDVH, SDVN, SDFB, SDFH, SDFI, SKMB, SKMN,
    n_heads, seqlen_q, seqlen_k, headdim, rotate_dim, HAS_POPE: tl.constexpr, IS_CAUSAL: tl.constexpr, HAS_MASK: tl.constexpr, BLOCK_HEADDIM: tl.constexpr,
    EVEN_M: tl.constexpr, EVEN_N: tl.constexpr, EVEN_HEADDIM: tl.constexpr, BM: tl.constexpr, BN: tl.constexpr,
):
    hb = tl.program_id(1)
    b, h = hb // n_heads, hb % n_heads
    sn = tl.program_id(0)

    om, on, od = tl.arange(0, BM), sn * BN + tl.arange(0, BN), tl.arange(0, BLOCK_HEADDIM)
    mn, md, mr = on < seqlen_k, od < headdim, od < rotate_dim

    k = tl.load(K + b * SKB + h * SKH + on[:, None] * SKN + od[None, :], mask = mn[:, None] & md[None, :], other = 0.0)
    v = tl.load(V + b * SVB + h * SVH + on[:, None] * SVN + od[None, :], mask = mn[:, None] & md[None, :], other = 0.0)

    if HAS_POPE:
        pk = tl.where(mr[None, :], tl.where(k > 20., k, tl.log(1.0 + tl.exp(k.to(tl.float32))).to(k.dtype)), k)
        fk = tl.load(Freqs + b * SFB + h * SFH + on[:, None] * SFI + od[None, :], mask = mn[:, None] & mr[None, :], other = 0.0)
        pb = tl.load(PopeBias + h * SPBH + od, mask = mr, other = 0.0)
        tk = (fk + pb[None, :]).to(tl.float32)
        kc, ks = tl.where(mr[None, :], pk * tl.cos(tk).to(k.dtype), pk), tl.where(mr[None, :], pk * tl.sin(tk).to(k.dtype), 0.0)
    else:
        kc, ks = k, None

    dv, dk = tl.zeros([BN, BLOCK_HEADDIM], tl.float32), tl.zeros([BN, BLOCK_HEADDIM], tl.float32)
    q_off = seqlen_k - seqlen_q

    for sm in range(0, seqlen_q, BM):
        m = sm + om
        mm = m < seqlen_q

        q = tl.load(Q + b * SQB + h * SQH + m[:, None] * SQM + od[None, :], mask = mm[:, None] & md[None, :], other = 0.0)

        if HAS_POPE:
            pq = tl.where(mr[None, :], tl.where(q > 20., q, tl.log(1.0 + tl.exp(q.to(tl.float32))).to(q.dtype)), q)
            fq = tl.load(Freqs + b * SFB + h * SFH + (q_off + m[:, None]) * SFI + od[None, :], mask = mm[:, None] & mr[None, :], other = 0.0).to(tl.float32)
            qc, qs = tl.where(mr[None, :], pq * tl.cos(fq).to(q.dtype), pq), tl.where(mr[None, :], pq * tl.sin(fq).to(q.dtype), 0.0)
            qk = tl.dot(qc, tl.trans(kc)) + tl.dot(qs, tl.trans(ks))
        else:
            qk = tl.dot(q, tl.trans(k))

        qk *= softmax_scale

        if IS_CAUSAL:
            qk += tl.where(m[:, None] >= on[None, :], 0, float("-inf"))

        if HAS_MASK:
            mask = tl.load(Mask + b * SKMB + on * SKMN, mask = mn, other = False)
            qk += tl.where(mask[None, :], 0, float("-inf"))

        lse = tl.load(Lse + hb * seqlen_q + m, mask = mm, other = float("-inf"))
        p = tl.exp(qk - tl.where(lse == float("-inf"), 0.0, lse)[:, None])
        p = tl.where((lse[:, None] == float("-inf")) | (~mm[:, None]), 0.0, p)

        do = tl.load(DO + b * SDB + h * SDH + m[:, None] * SDM + od[None, :], mask = mm[:, None] & md[None, :], other = 0.0)
        dv += tl.dot(tl.trans(p.to(do.dtype)), do)
        dp = tl.dot(do.to(p.dtype), tl.trans(v.to(p.dtype)))
        
        delta = tl.load(Delta + hb * seqlen_q + m, mask = mm, other = 0.0)
        ds = p * (dp - delta[:, None]) * softmax_scale

        if HAS_POPE:
            dqkc, dqks = tl.dot(ds.to(qc.dtype), kc), tl.dot(ds.to(ks.dtype), ks)
            dq = tl.where(mr[None, :], (dqkc * tl.cos(fq).to(dqkc.dtype) + dqks * tl.sin(fq).to(dqks.dtype)) * tl.sigmoid(q.to(tl.float32)).to(q.dtype), dqkc)
            dkkc, dkks = tl.dot(tl.trans(ds.to(qc.dtype)), qc), tl.dot(tl.trans(ds.to(qs.dtype)), qs)
            dk += tl.where(mr[None, :], (dkkc * tl.cos(tk).to(dkkc.dtype) + dkks * tl.sin(tk).to(dkks.dtype)) * tl.sigmoid(k.to(tl.float32)).to(k.dtype), dkkc)
            dfq = (dqks.to(tl.float32) * qc.to(tl.float32) - dqkc.to(tl.float32) * qs.to(tl.float32)).to(DFreqs.dtype.element_ty)
            tl.atomic_add(DFreqs + b * SDFB + h * SDFH + (q_off + m[:, None]) * SDFI + od[None, :], dfq, mask = mm[:, None] & mr[None, :])
            dfk = (dkks.to(tl.float32) * kc.to(tl.float32) - dkkc.to(tl.float32) * ks.to(tl.float32)).to(DFreqs.dtype.element_ty)
            tl.atomic_add(DFreqs + b * SDFB + h * SDFH + on[:, None] * SDFI + od[None, :], dfk, mask = mn[:, None] & mr[None, :])
            tl.atomic_add(DPopeBias + h * SPBH + od, tl.sum(dfk, 0), mask = mr)
        else:
            dq = tl.dot(ds.to(k.dtype), k)
            dk += tl.dot(tl.trans(ds.to(q.dtype)), q)

        tl.atomic_add(DQ + b * SDQB + h * SDQH + m[:, None] * SDQM + od[None, :], dq.to(DQ.dtype.element_ty), mask = mm[:, None] & md[None, :])

    tl.store(DV + b * SDVB + h * SDVH + on[:, None] * SDVN + od[None, :], dv.to(DV.dtype.element_ty), mask = mn[:, None] & md[None, :])
    tl.store(DK + b * SDKB + h * SDKH + on[:, None] * SDKN + od[None, :], dk.to(DK.dtype.element_ty), mask = mn[:, None] & md[None, :])

# wrapper functions

def flash_attn_forward(
    q, k, v,
    freqs = None,
    pope_bias = None,
    mask = None,
    causal = False,
    softmax_scale = None
):
    batch, seq_q, heads, d = q.shape
    seq_k = k.shape[1]
    softmax_scale = default(softmax_scale, d ** -0.5)

    has_p = exists(freqs) and exists(pope_bias)
    rot = freqs.shape[-1] if has_p else 0

    if has_p:
        if freqs.ndim == 2:
            fs = (0, 0, freqs.stride(0))
        elif freqs.ndim == 3:
            fs = (freqs.stride(0), 0, freqs.stride(1))
        else:
            fs = (freqs.stride(0), freqs.stride(2), freqs.stride(1))
    else:
        fs = (0, 0, 0)

    ps = pope_bias.stride(0) if has_p else 0

    lse = torch.empty((batch, heads, seq_q), device = q.device, dtype = torch.float32)
    o = torch.empty_like(q)

    BD = max(triton.next_power_of_2(d), 16)
    BM, BN = (64, 32) if d <= 64 else (32, 32)
    num_warps = 2 if d <= 32 else 4

    has_mask = exists(mask)
    if has_mask:
        skmb, skmn = mask.stride(0), mask.stride(1)
    else:
        skmb, skmn = 0, 0

    _fwd_kernel[(triton.cdiv(seq_q, BM), batch * heads)](
        q, k, v, freqs, pope_bias, o, lse, mask,
        softmax_scale,
        q.stride(0), q.stride(2), q.stride(1), k.stride(0), k.stride(2), k.stride(1), v.stride(0), v.stride(2), v.stride(1),
        *fs, ps, o.stride(0), o.stride(2), o.stride(1), skmb, skmn,
        heads, seq_q, seq_k, d, rot, has_p, causal, has_mask, BD, 
        EVEN_M=True, EVEN_N=True, EVEN_HEADDIM=True, BM=BM, BN=BN, num_warps=num_warps, num_stages=1
    )

    return o, lse

def flash_attn_backward(
    do, q, k, v, o, lse,
    dq, dk, dv,
    dfreqs = None,
    dpope_bias = None,
    freqs = None,
    pope_bias = None,
    mask = None,
    causal = False,
    softmax_scale = None
):
    batch, seq_q, heads, d = q.shape
    seq_k = k.shape[1]
    softmax_scale = default(softmax_scale, d ** -0.5)

    delta = torch.empty_like(lse)
    BD = max(triton.next_power_of_2(d), 16)

    _bwd_preprocess[(triton.cdiv(seq_q, 32), batch * heads)](
        o, do, delta,
        o.stride(0), o.stride(2), o.stride(1),
        do.stride(0), do.stride(2), do.stride(1),
        heads, seq_q, d, 32, BD
    )

    has_p = exists(freqs) and exists(pope_bias)
    rot = freqs.shape[-1] if has_p else 0

    if has_p:
        if freqs.ndim == 2:
            fs = (0, 0, freqs.stride(0))
        elif freqs.ndim == 3:
            fs = (freqs.stride(0), 0, freqs.stride(1))
        else:
            fs = (freqs.stride(0), freqs.stride(2), freqs.stride(1))

        if exists(dfreqs):
            if dfreqs.ndim == 2:
                dfs = (0, 0, dfreqs.stride(0))
            elif dfreqs.ndim == 3:
                dfs = (dfreqs.stride(0), 0, dfreqs.stride(1))
            else:
                dfs = (dfreqs.stride(0), dfreqs.stride(2), dfreqs.stride(1))
        else:
            dfs = (0, 0, 0)
    else:
        fs, dfs = (0, 0, 0), (0, 0, 0)

    ps = pope_bias.stride(0) if has_p else 0

    has_mask = exists(mask)
    if has_mask:
        skmb, skmn = mask.stride(0), mask.stride(1)
    else:
        skmb, skmn = 0, 0

    _bwd_kernel[(triton.cdiv(seq_k, 32), batch * heads)](
        q, k, v, freqs, pope_bias, do, dq, dk, dv, dfreqs, dpope_bias, lse, delta, mask, softmax_scale,
        q.stride(0), q.stride(2), q.stride(1), k.stride(0), k.stride(2), k.stride(1), v.stride(0), v.stride(2), v.stride(1),
        *fs, ps, do.stride(0), do.stride(2), do.stride(1), dq.stride(0), dq.stride(2), dq.stride(1),
        dk.stride(0), dk.stride(2), dk.stride(1), dv.stride(0), dv.stride(2), dv.stride(1), *dfs, skmb, skmn,
        heads, seq_q, seq_k, d, rot, has_p, causal, has_mask, BD, BM=32, BN=32, num_warps=2 if d <= 32 else 4, num_stages=1
    )

class FlashAttnFunction(Function):
    @staticmethod
    def forward(
        ctx,
        q, k, v,
        freqs = None,
        pope_bias = None,
        mask = None,
        causal = False,
        softmax_scale = None
    ):
        o, lse = flash_attn_forward(q, k, v, freqs, pope_bias, mask, causal, softmax_scale)
        ctx.save_for_backward(q, k, v, freqs, pope_bias, mask, o, lse)
        ctx.causal, ctx.softmax_scale = causal, softmax_scale
        return o

    @staticmethod
    def backward(ctx, do):
        do = do.contiguous()
        q, k, v, f, pb, m, o, lse = ctx.saved_tensors

        dq, dk, dv = torch.zeros_like(q), torch.zeros_like(k), torch.zeros_like(v)
        df, dpb = torch.zeros_like(f) if exists(f) else None, torch.zeros_like(pb) if exists(pb) else None

        flash_attn_backward(do, q, k, v, o, lse, dq, dk, dv, df, dpb, f, pb, m, ctx.causal, ctx.softmax_scale)
        return dq, dk, dv, df, dpb, None, None, None

def flash_attn(
    q, k, v,
    freqs = None,
    pope_bias = None,
    mask = None,
    causal = False,
    softmax_scale = None
):
    q, k, v = map(lambda t: t.contiguous(), (q, k, v))
    return FlashAttnFunction.apply(q, k, v, freqs, pope_bias, mask, causal, softmax_scale)



from __future__ import annotations
from math import pi

import torch
from torch import arange, cat, stack, is_tensor, Tensor, meshgrid
from torch.nn import Module, Parameter, ParameterList
from torch.amp import autocast

from einops import einsum, rearrange

from PoPE_pytorch.pope import PolarEmbedReturn, apply_pope_to_qk

# helper functions

def exists(v):
    return v is not None

def default(v, d):
    return v if exists(v) else d

# Axial PoPE class

class AxialPoPE(Module):
    # convenience
    apply_pope_to_qk = staticmethod(apply_pope_to_qk)

    def __init__(
        self,
        dim,
        *,
        heads,
        axial_dims: tuple[int, ...] | None = None,
        theta = 10000,
        bias_uniform_init = False,
    ):
        super().__init__()
        self.dim = dim
        self.heads = heads

        if not exists(axial_dims):
            axial_dims = (dim,)

        self.axial_dims = axial_dims
        assert sum(axial_dims) == dim, f'sum of axial_dims {axial_dims} must be equal to dim {dim}'

        # inv freqs for each axial dimension

        self.inv_freqs = ParameterList()

        for axial_dim in axial_dims:
            inv_freqs = theta ** -(arange(axial_dim).float() / axial_dim)
            self.inv_freqs.append(Parameter(inv_freqs, requires_grad = False))

        # the learned bias on the keys

        self.bias = Parameter(torch.zeros(heads, dim))

        if bias_uniform_init:
            self.bias.uniform_(-2. * pi, 0.)

    @property
    def device(self):
        return self.bias.device

    @staticmethod
    def get_grid_positions(*dims, device = None):
        grid = meshgrid(*[arange(d, device = device).float() for d in dims], indexing = 'ij')
        return stack([g.flatten() for g in reversed(grid)], dim = -1)

    @autocast('cuda', enabled = False)
    def forward(
        self,
        pos_or_dims: Tensor | tuple[int, ...],
    ):
        # handle auto grid generation if tuple is passed

        if isinstance(pos_or_dims, tuple):
            pos = self.get_grid_positions(*pos_or_dims, device = self.device)
        else:
            pos = pos_or_dims

        # pos shape is (..., N) where N is len(axial_dims)

        assert pos.shape[-1] == len(self.axial_dims)

        all_freqs = []

        for i, inv_freqs in enumerate(self.inv_freqs):
            # pos_i shape is (...)

            pos_i = pos[..., i]

            # freqs_i shape is (..., axial_dim)

            freqs_i = einsum(pos_i, inv_freqs, '... , d -> ... d')
            all_freqs.append(freqs_i)

        # concat axial freqs

        freqs = cat(all_freqs, dim = -1)

        # the bias, with clamping

        bias = self.bias.clamp(-2. * pi, 0.)

        return PolarEmbedReturn(freqs, bias)
